/*
 * mod_watch.c for Apache 1.3
 *
 * Copyright 2000 by Anthony Howe.  All rights reserved.
 *
 *
 * LICENSE
 * -------
 *
 * This source distribution is made freely available and there is no charge
 * for its use, provided you retain this notice, disclaimers, author's
 * copyright, and credits.
 *
 *
 * DISCLAIMER
 * ----------
 *
 * THIS SOFTWARE IS PROVIDE "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO WAY SHALL THE
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBM_HOST
#define DBM_HOST		"127.0.0.1";
#endif

#ifndef DBM_PORT
#define DBM_PORT		9866;
#endif

#ifndef DB_NAME
#define DB_NAME			"mod_watch";
#endif

#ifndef DB_SECRET
#define DB_SECRET		"public";
#endif

#ifndef DB_DIRECTORY
#define DB_DIRECTORY		"/var/spool/mod_watch/"
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define MODULE			"mod_watch"
#define AUTHOR			"achowe@snert.com"
#define VERSION			"2.1"
#define NDEBUG			1

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_config.h"

#include <pwd.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#define UNSET			-1
#define MERGE(p, c)		(c == UNSET ? p : c)
#define MERGE_PTR(p, c)		((void *) c == (void *) 0 ? p : c)

module MODULE_VAR_EXPORT watch_module;

typedef struct ml_socket {
	int fd;
	struct pool *pool;
	struct in_addr *ip;
	struct sockaddr_in address;
} ml_socket;

typedef struct dbm_connection {
	char *dbname;
	char *dbsecret;
	struct ml_socket *socket;
} dbm_connection;

typedef struct dbm_response {
	unsigned char version;
	unsigned char status;
	unsigned short length;
	unsigned char value[1];
} dbm_response;

/*
 *	counter32	ifInOctets	bytes received from POST requests
 *	counter32	ifOutOctets	bytes sent in response to requests
 *	counter32	ifRequests	request / hit counter
 *	counter32	ifDocument	document hit counter
 *	counter32	ifRefused	(future) request refused by policy
 *	integer		delay		(future) current response delay/credit
 *	integer		concurrent	(future) number of concurrent requests
 */
unsigned long ifInOctets, ifOutOctets, ifRequests, ifDocuments;

#define WATCH_SERVER		1
#define WATCH_VHOST		2
#define WATCH_OWNER		4
#define WATCH_ALL		(~0)

static int isWatchInfo = 0;
static int watch_what = WATCH_ALL;

static dbm_connection *dbm;
static long dbm_port = DBM_PORT;
static const char *dbm_host = DBM_HOST;
static const char *db_name = DB_NAME;
static const char *db_secret = DB_SECRET;
static const char *db_dir = DB_DIRECTORY;
static server_rec *main_server;

static array_header *document_conf;
static array_header *document_list;
static const char *documents_def[] = {
	"text/html",
	"text/plain",
	"text/xml",
	"application/pdf",
	"text/enriched",
	"text/richtext",
	"application/*word",
	"application/*excel",
	"application/postscript",
	"application/rtf",
/*	"httpd/unix-directory", */
	(char *) 0
};

typedef struct db_policy {
	const char *name;
	int (*logger) (request_rec *);
	int (*content) (request_rec *);
} db_policy;

static int log_dbm(request_rec *r);
static int log_files(request_rec *r);
static int log_random(request_rec *r);
static int log_shared(request_rec *r);

static int content_dbm(request_rec *r);
static int content_files(request_rec *r);
static int content_random(request_rec *r);
static int content_shared(request_rec *r);

static db_policy policy_table[] = {
	{ "dbm", log_dbm, content_dbm },
	{ "files", log_files, content_files },
	{ 0, 0 }
};

/***********************************************************************
 *** Support Routines
 ***********************************************************************/

#define UNKNOWN_UID		((uid_t) ~0)

/*
 */
static const char *
ml_id2uname(struct pool *p, uid_t uid)
{
	if (uid == UNKNOWN_UID)
		return "unknown-user";

#if defined(WIN32) || defined(NETWARE)
	return "default-user";
#else
{
	struct passwd *pw;

	if ((pw = getpwuid(uid)) != (struct passwd *) 0)
		return ap_pstrdup(p, pw->pw_name);

	return "unknown-user";
}
#endif
}

#define SECONDS_PER_MINUTE	60L
#define SECONDS_PER_HOUR	3600L
#define SECONDS_PER_DAY		86400L
#define SECONDS_PER_WEEK	(SECONDS_PER_DAY*7)
#define SECONDS_PER_MONTH	(SECONDS_PER_DAY*30)

static char *
ml_elapsed_time(struct pool * p, unsigned long s)
{
	int h, m;
	unsigned long d;

	d = s / SECONDS_PER_DAY;
	s %= SECONDS_PER_DAY;

	h = s / SECONDS_PER_HOUR;
	s %= SECONDS_PER_HOUR;

	m = s / SECONDS_PER_MINUTE;
	s %= SECONDS_PER_MINUTE;

	if (0 < d)
		return ap_psprintf(p, "%lu+%02d:%02d.%02d", d, h, m, (int) s);
	if (0 < h)
		return ap_psprintf(p, "%d:%02d.%02d", h, m, (int) s);
	if (0 < m)
		return ap_psprintf(p, "%d.%02d", m, (int) s);

	return ap_psprintf(p, "%d", (int) s);
}

/*
 * Convert ASCII IP address or host name into an in_addr structure.
 * Return a pointer to the structure or null if address is not found.
 */
static struct in_addr *
ml_socket_address(struct pool *p, const char *address)
{
	struct hostent *host;
	struct in_addr *saddr;

	saddr = ap_pcalloc(p, sizeof *saddr);

	/* First try it as aaa.bbb.ccc.ddd. */
	saddr->s_addr = inet_addr(address);
	if (saddr->s_addr != -1)
		return saddr;

	host = gethostbyname(address);
	if (host != (struct hostent *) 0) {
		*saddr = *(struct in_addr *) *host->h_addr_list;
		return saddr;
	}

	return (struct in_addr *) 0;
}

static struct ml_socket *
ml_socket_open(struct pool *p, const char *host, int port, int type, int proto)
{
	struct ml_socket target, *sock;

	/* Convert DBM server host address from ASCII to binary. */
	target.ip = ml_socket_address(p, host);
	if (target.ip == (struct in_addr *) 0)
		return (ml_socket *) 0;

	target.address.sin_addr.s_addr = target.ip->s_addr;
	target.address.sin_port = htons(port);
	target.address.sin_family = AF_INET;

	/* Get us the socket. */
	target.pool = p;
	target.fd = ap_psocket(p, PF_INET, type, proto);
	if (target.fd < 0)
		return (ml_socket *) 0;

	/* Now create our socket structure for the caller. */
	sock = ap_pcalloc(p, sizeof *sock);
	if (sock != (ml_socket *) 0)
		*sock = target;

	return sock;
}

static int
ml_socket_close(struct ml_socket *sock)
{
	if (sock != (struct ml_socket *) 0)
		ap_pclosesocket(sock->pool, sock->fd);
}

static dbm_connection *
dbm_connect(struct pool *p, struct ml_socket *sock, const char *db, const char *secret)
{
	dbm_connection *dc = ap_pcalloc(p, sizeof *dc);

	dc->socket = sock;
	dc->dbname = ap_pstrdup(p, db);
	dc->dbsecret = ap_pstrdup(p, secret);

	if (connect(sock->fd, (struct sockaddr *) &sock->address, sizeof sock->address) < 0)
		return (dbm_connection *) 0;

	return dc;
}

static dbm_response *
dbm_request(struct pool *p, dbm_connection *dbm, int cmd, const char *key, const unsigned char *value)
{
	fd_set rin;
	struct timeval tv;
	dbm_response *response;
	unsigned char *request, *ptr;
	size_t length, dlen, slen, klen, vlen;

	errno = 0;

	if (key == (const char *) 0) {
		errno = EINVAL;
		return (dbm_response *) 0;
	}

	if (value == (const unsigned char *) 0)
		value = "";

	/* Build request message. */
	dlen = strlen(dbm->dbname) & 0xff;
	slen = strlen(dbm->dbsecret) & 0xff;
	klen = strlen(key) & 0xff;
	vlen = strlen(value) & 0xffff;

	ptr = request = ap_pcalloc(p, 7 + dlen + slen + klen + vlen);

	/* Request message version */
	*ptr++ = 0;

	/* Request command */
	*ptr++ = (unsigned char) cmd;

	/* Request field lengths */
	*ptr++ = (unsigned char) dlen;
	*ptr++ = (unsigned char) slen;
	*ptr++ = (unsigned char) klen;
	*((unsigned short *) ptr)++ = (unsigned short) htons(vlen & 0xffff);

	/* Request fields */
	ptr = memcpy(ptr, dbm->dbname, dlen) + dlen;
	ptr = memcpy(ptr, dbm->dbsecret, slen) + slen;
	ptr = memcpy(ptr, key, klen) + klen;
	ptr = memcpy(ptr, value, vlen) + vlen;

	/* Send request. */
	length = write(dbm->socket->fd, request, ptr - request);

	if (length < ptr - request) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, main_server,
			"Error sending request to DBM server"
		);
		return (dbm_response *) 0;
	}

	/* Wait for a response. */
	FD_ZERO(&rin);
	FD_SET(dbm->socket->fd, &rin);
	tv.tv_sec = 30;
	tv.tv_usec = 0;

	if (select(dbm->socket->fd+1, &rin, (fd_set *) 0, (fd_set *) 0, &tv) == 0) {
#ifndef NDEBUG
		ap_log_error(
			APLOG_MARK, APLOG_DEBUG, main_server,
			"select()", dbm->socket->fd
		);
#endif
		return (dbm_response *) 0;
	}

	if ((length = read(dbm->socket->fd, request, 4)) != 4) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, main_server,
			"Error receiving response header from DBM server"
		);
		return (dbm_response *) 0;
	}

	if (request[0] != 0) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, main_server,
			"Unknown response message format"
		);
		return (dbm_response *) 0;
	}

	vlen = ntohs(*(unsigned short *)(request+2));
	response = (struct dbm_response *) ap_pcalloc(p, sizeof *response + vlen);
	response->status = request[1];
	response->length = (unsigned short) vlen;

	if ((length = read(dbm->socket->fd, response->value, vlen)) != vlen) {
		ap_log_error(
			APLOG_MARK, APLOG_ERR, main_server,
			"Error receiving response from DBM server"
		);
		return (dbm_response *) 0;
	}

	return response;
}

/***********************************************************************
 ***  Create, Init, and Merge Routines
 ***********************************************************************/

/*
 *
 */
static void
init_module(server_rec *s, struct pool *p)
{
	/* Handy to have for error logging. */
	main_server = s;

        ap_add_version_component(MODULE"/"VERSION);

	/* If no WatchDocuments specified, use defaults. */
	if (document_conf == (array_header *) 0) {
		const char **list, **mime;

		document_conf = ap_make_array(
			p,
			sizeof documents_def / sizeof *documents_def,
			sizeof *documents_def
		);

		for (list = documents_def; *list != (char *) 0; ++list) {
			mime = (const char **) ap_push_array(document_conf);
			if (mime == (const char **) 0)
				exit(APEXIT_INIT);
			*mime = *list;
		}
	}

	/* Prepare for next restart. */
        document_list = document_conf;
        document_conf = (array_header *) 0;

#ifndef NDEBUG
	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
		"main_server=%s", s->server_hostname
	);
#endif
}

/***********************************************************************
 ***  Handlers
 ***********************************************************************/

/*
 * All of our content handlers return plain text.
 */
static int
fixup_handler(request_rec *r)
{
	isWatchInfo = r->handler != (char *) 0
		&& ap_strcmp_match(r->handler, "watch-info") == 0;

	if (isWatchInfo) {
		r->content_type = "text/plain";
		return OK;
	}

	return DECLINED;
}

static void
counters(request_rec *r)
{
	const char *value;
	unsigned long content_length;

	ifRequests += 1;

	value = ap_table_get(r->headers_in, "Content-Length");
	if (value != (const char *) 0)
		ifInOctets += strtol(value, (char **) 0, 10);

	ifOutOctets += r->bytes_sent;

	/* Update ifDocuments counter. */
	if (r->content_type != (const char *) 0) {
		int i;
		request_rec *rr;
		char **list = (char **) document_list->elts;

		/* Find actual content handler issued. */
		for (rr = r; rr->next != (request_rec *) 0; rr = rr->next)
			;

		for (i = 0; i < document_list->nelts; ++i) {
			if (ap_strcasecmp_match(rr->content_type, list[i]) == 0) {
				ifDocuments += 1;
				break;
			}
		}
	}
}

/*
 * Pick a target, which is either the server name of the request
 * or a specified ~user name or ~SERVER.
 */
static const char *
target(request_rec *r)
{
	if (r->uri[0] == '/' && r->uri[1] == '~' && ap_isalnum(r->uri[2])) {
		const char *remainder = r->uri + 2;
		return ap_getword(r->pool, &remainder, '/');
	}

	return ap_get_server_name(r);
}

/*
 * Write the content, which is single text line containing the
 * target name, server uptime in seconds, octets in, octets out,
 * requests in, and documents out.
 */
static int
content_print(request_rec *r, const char *name)
{
	ap_send_http_header(r);

	if (r->header_only)
		return OK;

	ap_rprintf(
		r, "%s %lu %lu %lu %lu %lu\n", name,
		r->request_time - ap_restart_time,
		ifInOctets, ifOutOctets, ifRequests, ifDocuments
	);

	return OK;
}

/***********************************************************************
 ***  Log & Content using DBM Server
 ***********************************************************************/

static void
update_dbm(request_rec *r, const char *key)
{
	const char *value;
	dbm_response *response;

	/* Get the current counter values from the database. */
	response = dbm_request(r->pool, dbm, 'g', key, (char *) 0);
	if (response == (dbm_response *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_WARNING, r->server,
			"No response from DBM server (get %s)", key
		);
		return;
	}

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;
	(void) sscanf(
		response->value, "%lu %lu %lu %lu",
		&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
	);

	counters(r);

	/* Update the database for this key. */
	value = ap_psprintf(
		r->pool, "%lu %lu %lu %lu %s",
		ifInOctets, ifOutOctets, ifRequests, ifDocuments
	);

	response = dbm_request(r->pool, dbm, 's', key, value);
	if (response == (dbm_response *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_WARNING, r->server,
			"No response from DBM server (set %s)", key
		);
		return;
	}
}

/*
 * Watch the total number of bytes in & out for each request. Count
 * the values against the server, virtual host, and file owner.
 *
 * We do not bother with <directory> or <location> blocks, since
 * these will be referenced through a server name, which is a more
 * interesting and useful identifier for a customer.
 *
 * We record the data for all requests. No need to configure individual
 * virtual hosts or add specific user directives. The advantage of this
 * means we do not have to do anything to support mod_vhost_alias.
 * Everything just happens :)
 *
 * For those people using mod_vhost_alias, this module, and MRTG, a
 * special script is required to create the MRTG configuration based on
 * the mod_vhost_alias directives and the existing web directories.
 *
 * A similar issue exists for local users. The /etc/passwd file would
 * have to be parsed to generate an MRTG configuration for every user.
 * An MRTG entry for every user is probably overkill and hard to
 * maintain; better only to monitor specific users.
 */
static int
log_dbm(request_rec *r)
{
	struct ml_socket *sock;
	const char *key, *value;

	/* Do not count the watch-info reports. */
	if (isWatchInfo)
		return DECLINED;

	/* Find last sub-request for bytes actually sent. */
	for ( ; r->next != (request_rec *) 0; r = r->next)
		;

	/* We create the socket and DBM connection per request, because
	 * the DBM server could disappear on us. Note the connection will
	 * be closed by Apache when the request is completed.
	 */
	sock = ml_socket_open(r->pool, dbm_host, dbm_port, SOCK_STREAM, IPPROTO_TCP);
	if (sock == (ml_socket *) 0) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "Cannot reach DBM server");
		return DECLINED;
	}

	dbm = dbm_connect(r->pool, sock, db_name, db_secret);
	if (dbm == (dbm_connection *) 0) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "Cannot connect to DBM server");
		return DECLINED;
	}

	/* Update the uptime database entry. */
	value = ml_elapsed_time(r->pool, r->request_time - ap_restart_time);
	(void) dbm_request(r->pool, dbm, 's', "UPTIME", value);

	if (watch_what & WATCH_VHOST)
		update_dbm(r, ml_id2uname(r->pool, r->finfo.st_uid));

	if (watch_what & WATCH_OWNER)
		update_dbm(r, ap_get_server_name(r));

	if (watch_what & WATCH_SERVER)
		update_dbm(r, "SERVER");

	return DECLINED;
}

static int
content_dbm(request_rec *r)
{
	int rc;
	struct ml_socket *sock;
	dbm_response *response;
	const char *name = target(r);

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	/* We create the socket and DBM connection per request, because
	 * the DBM server could disappear on us. Note the connection will
	 * be closed by Apache when the request is completed.
	 */
	sock = ml_socket_open(r->pool, dbm_host, dbm_port, SOCK_STREAM, IPPROTO_TCP);
	if (sock == (ml_socket *) 0) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "Cannot reach DBM server");
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	dbm = dbm_connect(r->pool, sock, db_name, db_secret);
	if (dbm == (dbm_connection *) 0) {
		ap_log_error(APLOG_MARK, APLOG_ERR, r->server, "Cannot connect to DBM server");
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	response = dbm_request(r->pool, dbm, 'g', name, (char *) 0);
	if (response == (dbm_response *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_WARNING, r->server,
			"No response from DBM server (get %s)", name
		);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;
	(void) sscanf(
		response->value, "%lu %lu %lu %lu",
		&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
	);

	return content_print(r, name);
}

/***********************************************************************
 ***  Log & Content using Weenie Files
 ***********************************************************************/

static int
update_file(request_rec *r, const char *file)
{
	int fd;
	FILE *fp;

	if ((file = ap_pstrcat(r->pool, db_dir, file, (char *) 0)) == (char *) 0)
		return;

	if ((fd = ap_popenf(r->pool, file, O_CREAT|O_RDWR, 0600)) < 0)
		return;

	if ((fp = ap_pfdopen(r->pool, fd, "r+")) == (FILE *) 0)
		return;

	for (errno = 0; flock(fd, LOCK_EX) < 0; ) {
		if (errno != EINTR) {
			ap_pfclose(r->pool, fp);
			return;
		}
	}

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;
	(void) fscanf(
		fp, "%lu %lu %lu %lu",
		&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
	);

	counters(r);

	rewind(fp);
	(void) fprintf(
		fp, "%lu %lu %lu %lu\n",
		ifInOctets, ifOutOctets, ifRequests, ifDocuments
	);

	ap_pfclose(r->pool, fp);
}

static int
log_files(request_rec *r)
{
	/* Do not count the watch-info reports. */
	if (isWatchInfo)
		return DECLINED;

	/* Find last sub-request for bytes actually sent. */
	for ( ; r->next != (request_rec *) 0; r = r->next)
		;

	if (watch_what & WATCH_VHOST)
		update_file(r, ml_id2uname(r->pool, r->finfo.st_uid));

	if (watch_what & WATCH_OWNER)
		update_file(r, ap_get_server_name(r));

	if (watch_what & WATCH_SERVER)
		update_file(r, "SERVER");

	return DECLINED;
}

static int
content_files(request_rec *r)
{
	int rc;
	FILE *fp;
	const char *file;
	const char *name = target(r);

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;

	file = ap_pstrcat(r->pool, db_dir, name, (char *) 0);
	if ((fp = ap_pfopen(r->pool, file, "r")) != (FILE *) 0) {
		(void) fscanf(
			fp, "%lu %lu %lu %lu",
			&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
		);
		ap_pfclose(r->pool, fp);
	}

	return content_print(r, name);
}

/***********************************************************************
 ***  Configuration Directives
 ***********************************************************************/

/*
 * Allows the URL used to query virtual host data:
 *
 *	http://www.snert.com/watch-info
 *
 *	<Location /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 * Allows the URL used to query file owner and web
 * server data:
 *
 *	http://www.snert.com/~achowe/watch-info
 *	http://www.snert.com/~SERVER/watch-info
 *
 *	<Location "/~* /watch-info>"
 *	SetHandler watch-info
 *	</Location>
 */
static handler_rec content_handlers[] = {
	{ "watch-info", content_files },
	{ NULL }
};

/*
 * WatchStore dbm [host port [dbname dbsecret]]
 * WatchStore files [directory]
 */
static const char *
WatchStore(cmd_parms *cmd, void *dconf, const char *args)
{
	long value;
	db_policy *dp;
	char *policy, *arg;

	/* First argument is required. */
	if ((policy = ap_getword_white(cmd->pool, &args)) == (char *) 0)
		return "Storage policy not specified.";

	for (dp = policy_table; dp->name != (const char *) 0; ++dp) {
		if (ap_strcasecmp_match(policy, dp->name) == 0)
			break;
	}

	switch (dp - policy_table) {
	case 0:
		/* WatchStore dbm [host port [dbname dbsecret]] */
		if ((arg = ap_getword_white(cmd->pool, &args)) == (char *) 0)
			break;

		dbm_host = arg;

		if ((arg = ap_getword_white(cmd->pool, &args)) == (char *) 0)
			return "Missing DBM server port.";

		value = strtol(arg, (char **) 0, 10);

		if (value < 0)
			dbm_port = 0;
		else if (UINT_MAX < value)
			dbm_port = UINT_MAX;
		else
			dbm_port = (unsigned int) value;

		if ((arg = ap_getword_white(cmd->pool, &args)) == (char *) 0)
			break;

		db_name = arg;

		if ((arg = ap_getword_white(cmd->pool, &args)) == (char *) 0)
			return "Missing DBM secret.";

		db_secret = arg;
		break;
	case 1:
		/* WatchStore files [directory] */
		if ((arg = ap_getword_white(cmd->pool, &args)) != (char *) 0) {
			struct stat buf;
			if (*arg != '/')
				return "Directory pathname must be absolute.";

			if (stat(arg, &buf) != 0 || !S_ISDIR(buf.st_mode))
				return "Pathname must be a directory.";

			/* Append trailing slash. */
			if (strrchr(arg, '/')[1] != '\0')
				arg = ap_pstrcat(cmd->pool, arg, "/", (char *) 0);

			db_dir = arg;
		}
		break;
	default:
		return "Invalid storage policy.";
	}

	/* Set the handler to be used for logging. */
	watch_module.logger = dp->logger;
	content_handlers[0].handler = dp->content;

	return (const char *) 0;
}

/*
 * WatchDocuments mime-type ...
 */
static const char *
WatchDocuments(cmd_parms *cmd, void *dconf, char *arg)
{
	char **mime;

	if (document_conf == (array_header *) 0)
		document_conf = ap_make_array(cmd->pool, 0, sizeof *documents_def);

	mime = (char **) ap_push_array(document_conf);
	if (mime == (char **) 0)
		return "Cannot add to WatchDocuments list.";

	*mime = (char *) ap_pstrdup(cmd->pool, arg);

	return (const char *) 0;
}

/*
 * WatchWhat item ...
 */
static const char *
WatchWhat(cmd_parms *cmd, void *dconf, char *arg)
{
	if (ap_strcasecmp_match(arg, "virtual-host") == 0)
		watch_what |= WATCH_VHOST;
	else if (ap_strcasecmp_match(arg, "file-owner") == 0)
		watch_what |= WATCH_OWNER;
	else if (ap_strcasecmp_match(arg, "server") == 0)
		watch_what |= WATCH_SERVER;
	else if (ap_strcasecmp_match(arg, "all") == 0)
		watch_what |= WATCH_VHOST|WATCH_OWNER|WATCH_SERVER;
	else
		return "Invalid item for WatchWhat";

	return (const char *) 0;
}

static command_rec command_table[] = {
	{ "WatchDocuments", WatchDocuments, NULL, RSRC_CONF, ITERATE,
	"List of MIME types considered to be documents." },

	{ "WatchStore", WatchStore, NULL, RSRC_CONF, RAW_ARGS,
	"The storage policy: dbm [host port [dbname dbsecret]] ; files [directory]" },

	{ "WatchWhat", WatchWhat, NULL, RSRC_CONF, ITERATE,
	"One or more items to watch: all, virtual-host, file-owner, server." },

	{ NULL }
};

/***********************************************************************
 ***  Module Definition Block
 ***********************************************************************/

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT watch_module = {
	STANDARD_MODULE_STUFF,
	init_module,		/* module initializer                  */
	NULL,			/* create per-dir    conf structures */
	NULL,			/* merge  per-dir    conf structures */
	NULL,			/* create per-server conf structures */
	NULL,			/* merge  per-server conf structures */
	command_table,		/* table of conf file commands       */
	content_handlers,	/* [#8] MIME-typed-dispatched handlers */
	NULL,			/* [#1] URI to filename translation    */
	NULL,			/* [#4] validate user id from request  */
	NULL,			/* [#5] check if the user is ok _here_ */
	NULL,			/* [#3] check access by host address   */
	NULL,			/* [#6] determine MIME type            */
	fixup_handler,		/* [#7] pre-run fixups                 */
	log_files,		/* [#9] log a transaction              */
#if MODULE_MAGIC_NUMBER >= 19970103
	NULL,			/* [#2] header parser                  */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
	NULL,			/* child_init			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
	NULL,			/* child_exit			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	NULL			/* [#0] post read-request              */
#endif
};

